//------------------------------------------------------------------------------
//SFWK Framework
//Copyright (C) 2011 SQLI

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------

using System;

namespace SFWK.Core.Unity.DynamicProxy
{
	/// <summary>
	/// Factory for creating Dynamic proxy instances
	/// <code>
	/// TestClasses.SimpleClass testClass = new TestClasses.SimpleClass();
	/// TestClasses.ISimpleInterface testClassProxy = (TestClasses.ISimpleInterface) DynamicProxyFactory.Instance.CreateProxy(testClass, new InvocationDelegate(InvocationHandler));
	/// testClassProxy.Method1();
	/// </code>
	/// <see cref="IDynamicProxy"/>
	/// </summary>
	public class DynamicProxyFactory {
		private static DynamicProxyFactory self = new DynamicProxyFactory();

		private DynamicProxyFactory() {
		}

		/// <summary>
		/// Get the instance of the factory (singleton)
		/// </summary>
		public static DynamicProxyFactory Instance {
			get { return self; }
		}

		/// <summary>
		/// Create a proxy for the target object
		/// </summary>
		/// <param name="target">The object to create a proxy for</param>
		/// <param name="invocationHandler">The invocation handler for the proxy</param>
		/// <returns>The dynamic proxy instance</returns>
		public object CreateProxy(object target, InvocationDelegate invocationHandler) {
			return CreateProxy(target, invocationHandler, false, null);
		}

		/// <summary>
		/// Create a proxy for the target object
		/// </summary>
		/// <param name="target">The object to create a proxy for</param>
		/// <param name="invocationHandler">The invocation handler for the proxy</param>
		/// <param name="strict">Indicates if the cast support should be strict. If strict is true all casts are checked before being performed</param>
		/// <returns>The dynamic proxy instance</returns>
		public object CreateProxy(object target, InvocationDelegate invocationHandler, bool strict) {
			return CreateProxy(target, invocationHandler, strict, null);
		}

		/// <summary>
		/// Create a proxy for the target object
		/// </summary>
		/// <param name="target">The object to create a proxy for</param>
		/// <param name="invocationHandler">The invocation handler for the proxy</param>
		/// <param name="strict">Indicates if the cast support should be strict. If strict is true all casts are checked before being performed. The supportedType list will enabled support for more interfaces than the target object supports</param>
		/// <param name="supportedTypes">List of types that are supported for casts. Is only checked if strict is true.</param>
		/// <returns>The dynamic proxy instance</returns>
		public object CreateProxy(object target, InvocationDelegate invocationHandler, bool strict, Type[] supportedTypes) {
			return new DynamicProxyImpl(target, invocationHandler, strict, supportedTypes).GetTransparentProxy();
		}
	}
}

